/*
 * ruby interface for libxml <http://xmlsoft.org>
 */

#include <stdio.h>

#include <ruby.h>

#include <libxml/xmlmemory.h>
#include <libxml/parser.h>
#include <libxml/parserInternals.h>
#include <libxml/HTMLparser.h>
#include <libxml/HTMLtree.h>
#include <libxml/tree.h>
#include <libxml/xpath.h>
#include <libxml/debugXML.h>

#ifndef xmlChildreNode
# define xmlChildrenNode children
#endif


VALUE cXML;

VALUE cXMLNode;

char *xmlNodeTypeString[] =
{
    "NONE",
    "ELEMENT_NODE",
    "ATTRIBUTE_NODE",
    "TEXT_NODE",
    "CDATA_SECTION_NODE",
    "ENTITY_REF_NODE",
    "ENTITY_NODE",
    "PI_NODE",
    "COMMENT_NODE",
    "DOCUMENT_NODE",
    "DOCUMENT_TYPE_NODE",
    "DOCUMENT_FRAG_NODE",
    "NOTATION_NODE",
    "HTML_DOCUMENT_NODE",
    "DTD_NODE",
    "ELEMENT_DECL",
    "ATTRIBUTE_DECL",
    "ENTITY_DECL",
};

VALUE cXMLNs;



VALUE
xml_parse_string(klass, src)
VALUE klass, src;
{
    xmlDocPtr doc;

    doc = xmlParseMemory(RSTRING(src)->ptr, RSTRING(src)->len);
    if (doc == NULL)
    {
	rb_raise(rb_eRuntimeError, "parse failure");
    }

    return Data_Wrap_Struct(cXML, 0, xmlFreeDoc, doc);
}

VALUE
xml_parse_file(klass, src)
VALUE klass, src;
{
    xmlDocPtr doc;

    doc = xmlParseFile(RSTRING(src)->ptr);
    if (doc == NULL)
    {
	rb_raise(rb_eRuntimeError, "parse failure");
    }

    return Data_Wrap_Struct(cXML, 0, xmlFreeDoc, doc);
}

VALUE
xml_no_blanks(argc, argv, klass)
int argc;
VALUE *argv;
VALUE klass;
{
    VALUE flag;
    int fl = 0, ret;

    if (rb_scan_args(argc, argv, "01", &flag) == 0)
    {
	flag = Qtrue;
    }

    if (flag == Qnil || flag == Qfalse)
    {
	fl = 1;
    }
    ret = xmlKeepBlanksDefault(fl);

    return ret == 0 ? Qtrue : Qfalse;
}

VALUE
xml_no_entities(argc, argv, klass)
int argc;
VALUE *argv;
VALUE klass;
{
    VALUE flag;
    int fl = 1, ret;

    if (rb_scan_args(argc, argv, "01", &flag) == 0)
    {
	flag = Qtrue;
    }

    if (flag == Qnil || flag == Qfalse)
    {
	fl = 0;
    }
    ret = xmlSubstituteEntitiesDefault(fl);

    return ret == 1 ? Qtrue : Qfalse;
}


VALUE
xml_get_root(self)
VALUE self;
{
    xmlDocPtr doc;
    xmlNodePtr node;

    Data_Get_Struct(self, xmlDoc, doc);
    node = xmlDocGetRootElement(doc);

    return Data_Wrap_Struct(cXMLNode, 0, 0, node);
}

VALUE
xml_dump_string(self)
VALUE self;
{
    xmlDocPtr doc;
    xmlChar *ptr;
    int size;

    Data_Get_Struct(self, xmlDoc, doc);
    xmlDocDumpMemory(doc, &ptr, &size);
    if (size == 0)
    {
	return Qnil;
    }

    return rb_str_new(ptr, size);
}



VALUE
xmlnode_each_node(self)
VALUE self;
{
    xmlNodePtr ptr;
    VALUE node;

    Data_Get_Struct(self, xmlNode, ptr);
    ptr = ptr->xmlChildrenNode;

    while (ptr)
    {
	node = Data_Wrap_Struct(cXMLNode, 0, 0, ptr);
	rb_yield(node);
	ptr = ptr->next;
    }

    return Qnil;
}

VALUE
xmlnode_get_prop(self, name)
VALUE self, name;
{
    xmlNodePtr ptr;
    xmlChar *prop;

    Data_Get_Struct(self, xmlNode, ptr);
    prop = xmlGetProp(ptr, RSTRING(name)->ptr);
    if (prop == NULL)
    {
	return Qnil;
    }

    return rb_str_new2(prop);
}

VALUE
xmlnode_get_props(self)
VALUE self;
{
    xmlNodePtr ptr;
    xmlAttrPtr attr;
    VALUE props;

    Data_Get_Struct(self, xmlNode, ptr);
    attr = ptr->properties;

    props = rb_hash_new();

    while (attr)
    {
	rb_hash_aset(props, rb_str_new2(attr->name),
		     rb_str_new2(xmlNodeListGetString(ptr->doc,
						      attr->xmlChildrenNode,
						      1)));
	attr = attr->next;
    }

    return props;
}

VALUE
xmlnode_set_prop(self, name, val)
VALUE self, name, val;
{
    xmlNodePtr ptr;

    Data_Get_Struct(self, xmlNode, ptr);
    xmlSetProp(ptr, RSTRING(name)->ptr, RSTRING(val)->ptr);

    return self;
}

VALUE
xmlnode_name(self)
VALUE self;
{
    xmlNodePtr ptr;

    Data_Get_Struct(self, xmlNode, ptr);

    return rb_str_new2(ptr->name);
}

VALUE
xmlnode_set_name(self, name)
VALUE self, name;
{
    xmlNodePtr ptr;

    Data_Get_Struct(self, xmlNode, ptr);
    xmlNodeSetName(ptr, RSTRING(name)->ptr);

    return self;
}

VALUE
xmlnode_lang(self)
VALUE self;
{
    xmlNodePtr ptr;
    xmlChar *lang;

    Data_Get_Struct(self, xmlNode, ptr);
    lang = xmlNodeGetLang(ptr);
    if (lang == NULL)
    {
	return Qnil;
    }

    return rb_str_new2(lang);
}

VALUE
xmlnode_set_lang(self, lang)
VALUE self, lang;
{
    xmlNodePtr ptr;

    Data_Get_Struct(self, xmlNode, ptr);
    xmlNodeSetLang(ptr, RSTRING(lang)->ptr);

    return self;
}

VALUE
xmlnode_content(self)
VALUE self;
{
    xmlNodePtr ptr;
    xmlChar *str;

    Data_Get_Struct(self, xmlNode, ptr);
    str = xmlNodeGetContent(ptr);
    if (str == NULL)
    {
	return rb_str_new(0, 0);
    }

    return rb_str_new2(str);
}

VALUE
xmlnode_set_content(self, cont)
VALUE self, cont;
{
    xmlNodePtr ptr;

    Data_Get_Struct(self, xmlNode, ptr);
    xmlNodeSetContent(ptr, RSTRING(cont)->ptr);

    return self;
}

VALUE
xmlnode_add_content(self, cont)
VALUE self, cont;
{
    xmlNodePtr ptr;

    Data_Get_Struct(self, xmlNode, ptr);
    xmlNodeAddContent(ptr, RSTRING(cont)->ptr);

    return self;
}

VALUE
xmlnode_node_type(self)
VALUE self;
{
    xmlNodePtr ptr;

    Data_Get_Struct(self, xmlNode, ptr);

    return INT2FIX(ptr->type);
}

VALUE
xmlnode_is_text(self)
VALUE self;
{
    xmlNodePtr ptr;

    Data_Get_Struct(self, xmlNode, ptr);

    return xmlNodeIsText(ptr) ? Qtrue : Qfalse;
}

VALUE
xmlnode_is_blank(self)
VALUE self;
{
    xmlNodePtr ptr;

    Data_Get_Struct(self, xmlNode, ptr);

    return xmlIsBlankNode(ptr) ? Qtrue : Qfalse;
}


VALUE
xmlnode_prev(self)
VALUE self;
{
    xmlNodePtr ptr;

    Data_Get_Struct(self, xmlNode, ptr);

    return Data_Wrap_Struct(cXMLNode, 0, 0, ptr->prev);
}

VALUE
xmlnode_next(self)
VALUE self;
{
    xmlNodePtr ptr;

    Data_Get_Struct(self, xmlNode, ptr);

    return Data_Wrap_Struct(cXMLNode, 0, 0, ptr->next);
}

VALUE
xmlnode_parent(self)
VALUE self;
{
    xmlNodePtr ptr;

    Data_Get_Struct(self, xmlNode, ptr);

    return Data_Wrap_Struct(cXMLNode, 0, 0, ptr->parent);
}

VALUE
xmlnode_first_child(self)
VALUE self;
{
    xmlNodePtr ptr;

    Data_Get_Struct(self, xmlNode, ptr);

    return Data_Wrap_Struct(cXMLNode, 0, 0, ptr->children);
}

VALUE
xmlnode_children(self)
VALUE self;
{
    xmlNodePtr ptr;
    VALUE children;

    Data_Get_Struct(self, xmlNode, ptr);
    ptr = ptr->children;

    children = rb_ary_new();
    while (ptr)
    {
	rb_ary_push(children, Data_Wrap_Struct(cXMLNode, 0, 0, ptr));
	ptr = ptr->next;
    }

    return children;
}


VALUE
xmlnode_inspect(self)
VALUE self;
{
    xmlNodePtr ptr;
    char buf[128];

    Data_Get_Struct(self, xmlNode, ptr);
    snprintf(buf, 128-1, "<XML::XMLNode: name = \"%s\", type = \"%s\">",
	     ptr->name, xmlNodeTypeString[ptr->type]);

    return rb_str_new2(buf);
}



void
Init_xml()
{
    cXML = rb_define_class("XML", rb_cObject);

    rb_undef_method(CLASS_OF(cXML), "new");

    rb_define_singleton_method(cXML, "parse_string", xml_parse_string, 1);
    rb_define_singleton_method(cXML, "parse_file", xml_parse_file, 1);
    rb_define_singleton_method(cXML, "no_blanks", xml_no_blanks, -1);
    rb_define_singleton_method(cXML, "no_entities", xml_no_entities, -1);

    rb_define_method(cXML, "get_root", xml_get_root, 0);
    rb_define_method(cXML, "dump_string", xml_dump_string, 0);


    cXMLNode = rb_define_class_under(cXML, "XMLNode", rb_cObject);

    rb_define_const(cXMLNode, "ELEMENT_NODE", INT2FIX(1));
    rb_define_const(cXMLNode, "ATTRIBUTE_NODE", INT2FIX(2));
    rb_define_const(cXMLNode, "TEXT_NODE", INT2FIX(3));
    rb_define_const(cXMLNode, "CDATA_SECTION_NODE", INT2FIX(4));
    rb_define_const(cXMLNode, "ENTITY_REF_NODE", INT2FIX(5));
    rb_define_const(cXMLNode, "ENTITY_NODE", INT2FIX(6));
    rb_define_const(cXMLNode, "PI_NODE", INT2FIX(7));
    rb_define_const(cXMLNode, "COMMENT_NODE", INT2FIX(8));
    rb_define_const(cXMLNode, "DOCUMENT_NODE", INT2FIX(9));
    rb_define_const(cXMLNode, "DOCUMENT_TYPE_NODE", INT2FIX(10));
    rb_define_const(cXMLNode, "DOCUMENT_FRAG_NODE", INT2FIX(11));
    rb_define_const(cXMLNode, "NOTATION_NODE", INT2FIX(12));
    rb_define_const(cXMLNode, "HTML_DOCUMENT_NODE", INT2FIX(13));
    rb_define_const(cXMLNode, "DTD_NODE", INT2FIX(14));
    rb_define_const(cXMLNode, "ELEMENT_DECL", INT2FIX(15));
    rb_define_const(cXMLNode, "ATTRIBUTE_DECL", INT2FIX(16));
    rb_define_const(cXMLNode, "ENTITY_DECL", INT2FIX(17));

    rb_undef_method(CLASS_OF(cXMLNode), "new");

    rb_define_method(cXMLNode, "each_child", xmlnode_each_node, 0);
    rb_define_alias(cXMLNode, "each", "each_child");
    rb_define_method(cXMLNode, "get_prop", xmlnode_get_prop, 1);
    rb_define_method(cXMLNode, "get_props", xmlnode_get_props, 0);
    rb_define_method(cXMLNode, "set_prop", xmlnode_set_prop, 2);
    rb_define_method(cXMLNode, "name", xmlnode_name, 0);
    rb_define_method(cXMLNode, "name=", xmlnode_set_name, 1);
    rb_define_method(cXMLNode, "lang", xmlnode_lang, 0);
    rb_define_method(cXMLNode, "lang=", xmlnode_set_lang, 1);
    rb_define_method(cXMLNode, "content", xmlnode_content, 0);
    rb_define_method(cXMLNode, "content=", xmlnode_set_content, 1);
    rb_define_method(cXMLNode, "add_content", xmlnode_add_content, 1);
    rb_define_method(cXMLNode, "node_type", xmlnode_node_type, 0);
    rb_define_method(cXMLNode, "is_text?", xmlnode_is_text, 0);
    rb_define_method(cXMLNode, "is_blank?", xmlnode_is_blank, 0);

    rb_define_method(cXMLNode, "prev", xmlnode_prev, 0);
    rb_define_method(cXMLNode, "next", xmlnode_next, 0);
    rb_define_method(cXMLNode, "parent", xmlnode_parent, 0);
    rb_define_method(cXMLNode, "first_child", xmlnode_first_child, 0);
    rb_define_method(cXMLNode, "children", xmlnode_children, 0);

    rb_define_method(cXMLNode, "to_s", xmlnode_content, 0);
    rb_define_method(cXMLNode, "inspect", xmlnode_inspect, 0);


    cXMLNs = rb_define_class_under(cXML, "XMLNs", rb_cObject);

    return;
}
